==> nui_web-0.0.3.tar.gz: lib/nui_web.dart <==
library nui_web;
import 'dart:convert';
import 'dart:developer';
import 'dart:io';
import 'dart:typed_data';
import 'package:flutter/cupertino.dart';
import 'package:http/http.dart' as http;
import 'package:nui_core/busevent/nui_busevent.dart';
import 'package:nui_core/code/nui_code_util.dart';
import 'package:nui_core/data/nui_data_util.dart';
import 'package:nui_core/data/nui_ent.dart';
import 'package:nui_core/nui_core.dart';
import 'package:nui_error_handler/nui_error_handler.dart';
import 'package:nui_web/nui_connection.dart';
import 'nui_web_models.dart';

export 'package:nui_web/nui_web.dart';
export 'package:nui_web/nui_web_models.dart';
export 'package:nui_web/nui_connection.dart';

typedef HttpHeaderSetter = Map<String, String> Function();
typedef HttpExceptionWrapper = Future<NUIHttpResponse> Function();
typedef HttpAdditionalHandler = void Function(NUIHttpResponse);

abstract class NUIWebInterface{
  Future<NUIHttpResponse<T>> doPOST<T extends NUIEnt>({@required String url, Map<String, String> headers, NUIEnt bodyEnt, Map<String, dynamic> body, Encoding encoding, String pathVariable, NUIEntMapper<T> entity, bool storeTemp});
  Future<NUIHttpResponse<T>> doPUT<T extends NUIEnt>({@required String url, Map<String, String> headers, NUIEnt bodyEnt, Map<String, dynamic> body, Encoding encoding, String pathVariable, NUIEntMapper<T> entity, bool storeTemp});
  Future<NUIHttpResponse<T>> doPATCH<T extends NUIEnt>({@required String url, Map<String, String> headers, NUIEnt bodyEnt, Map<String, dynamic> body, Encoding encoding, String pathVariable, NUIEntMapper<T> entity, bool storeTemp});
  Future<NUIHttpResponse<T>> doDELETE<T extends NUIEnt>({@required String url, Map<String, String> headers, Encoding encoding, String pathVariable, NUIEntMapper<T> entity, bool storeTemp});
  Future<NUIHttpResponse<T>> doGET<T extends NUIEnt>({@required String url, Map<String, String> headers, String pathVariable, Map<String, String> params, NUIEntMapper<T> entity, bool storeTemp});
}

class NUIWeb implements NUIWebInterface{
  NUIWeb._builder(NUIWebBuilder builder) : _baseUrl = builder._baseUrl, _preHandler = builder._preHandler, _postHandlerDelay = builder._postHandlerDelay,  _postHandler = builder._postHandler, _headers = builder._headers, _statusMessages = builder._statusMessages ?? Map(), _errorHandler = builder._errorHandler, _connectivityListener = builder._connectivityListener, _module = isNullOrEmpty(builder._module) ? _defaultModule : builder._module{
    _webMap[_module] = this;
    NUISharedPrefBuilder().build();
    NUIBusEventBuilder().module(_getBusEventModule()).build();
    NUIConnectionUtil.subscribeConnectivityChange((result, hasInternet) {
      _getBusEvent().send<bool>(NUIBusEventData<bool>(data: hasInternet, sender: _getBusEventModule(), code: NUIConnectionUtil.CONNECTIVITY_CHANNEL), channel: NUIConnectionUtil.CONNECTIVITY_CHANNEL);
      if(_connectivityListener != null) {
        _connectivityListener(result, hasInternet);
      }
    });
  }

  final Map<int, String> _statusMessages;
  final String _module;
  final String _baseUrl;
  final HttpAdditionalHandler _preHandler;
  final HttpAdditionalHandler _postHandler;
  final Duration _postHandlerDelay;
  final ConnectivityChangeListener _connectivityListener;
  final HttpHeaderSetter _headers;
  final NUIErrorHandler _errorHandler;
  static Map<String, NUIWeb> _webMap = Map();
  static const String _defaultModule = "NUIWebDefaultModule";

  static Future<Uint8List> downloadFromURL(String url) async{
    try {
      http.Client client = new http.Client();
      var req = await client.get(Uri.parse(url));
      var bytes = req.bodyBytes;
      return bytes;
    }catch(e){
      return null;
    }
  }

  static NUIWeb get({String module}){
    NUIWeb instance = _webMap[isNullOrEmpty(module) ? _defaultModule : module];
    if(instance == null) throw Exception("NUIWeb -> NUIWeb instance has not been initialized with builder");
    return instance;
  }

  NUIBusEvent _getBusEvent(){
    return NUIBusEvent.get(module: _getBusEventModule());
  }

  void sendBusEvent<T>(NUIBusEventData<T> data, {String channel: NUIConnectionUtil.CONNECTIVITY_CHANNEL}){
    _getBusEvent().send<T>(data, channel: channel);
  }

  void subscribeBusEvent<T>(NUIBusEventListener listener, {String channel: NUIConnectionUtil.CONNECTIVITY_CHANNEL}){
    _getBusEvent().listen(listener, channel: channel);
  }

  void disposeBusEvent(NUIBusEventListener listener, {String channel: NUIConnectionUtil.CONNECTIVITY_CHANNEL}){
    _getBusEvent().dispose(listener, channel: channel);
  }

  String _getBusEventModule(){
    return "NUIWebBusEvent_" + (_module ?? _defaultModule);
  }

  String _getProcessedUrl(String url, String pathVariable, Map<String, String> params){
    String webUrl = isNullOrEmpty(_baseUrl) ? url : (_baseUrl[_baseUrl.length - 1] != '/' ? (_baseUrl + "/" + url) : (_baseUrl + url));
    if(!isNullOrEmpty(pathVariable)){
      if(webUrl[webUrl.length - 1] != '/') {
        webUrl = webUrl + "/" + pathVariable;
      }
      else{
        webUrl = webUrl + pathVariable;
      }
    }
    webUrl =  webUrl + _getPathParams(params);
    log("NUIWeb -> url : $webUrl");
    return webUrl;
  }

  void _logWeb(String message){
    log("NUIWeb -> $message");
  }

  void _preHandle<T>(NUIHttpResponse<T> response){
    if(_preHandler != null) _preHandler(response);
  }

  void _postHandle<T>(NUIHttpResponse<T> response){
    if(_postHandler != null){
      NUIAsync.delay(_postHandlerDelay, () {
        _postHandler(response);
      });
    }
  }

  String _getPathParams(Map<String, String> params){
    if(params == null || params.isEmpty){
      return "";
    }
    String pathParam = "";
    for(String key in params.keys){
      pathParam = pathParam + (pathParam.isEmpty ? "?$key=${params[key]}" : "&$key=${params[key]}");
    }
    return pathParam;
  }

  Map<String, String> _getHeaders(Map<String, String> additionalHeaders){
    if(_headers == null) return additionalHeaders;
    if(additionalHeaders == null) return _headers();
    Map<String, String> headers = _headers();
    for(String key in additionalHeaders.keys){
      headers[key] = additionalHeaders[key];
    }
    if(headers.isNotEmpty){
      _logWeb("Headers : " + headers.toString());
    }
    return headers;
  }

  Future<NUIHttpResponse<T>> wrapException<T extends NUIEnt>(HttpExceptionWrapper task) async{
    try {
      return task();
    }catch(e){
      _logWeb("Encountered an exception : ${e.toString()}");
      return NUIHttpResponse(success: false, code: HttpStatus.internalServerError, message: e.toString());
    }
  }

  Future<bool> _checkInternetConnection() async{
    return NUIConnectionUtil.hasConnection();
  }

  void _storeCacheResponse({String url, String pathVariable, http.Response response}){
    if(response.statusCode != HttpStatus.ok) return;
    final fullUrl = _getProcessedUrl(url, pathVariable, null);
    final cacheResponse = NUIHttpCacheResponse(response: response.body, headers: response.headers);
    NUISharedPref.get().save(fullUrl, cacheResponse);
  }

  NUIHttpCacheResponse getCacheResponse({String url, String pathVariable}){
    final fullUrl = _getProcessedUrl(url, pathVariable, null);
    return NUISharedPref.get().quickGetObject(fullUrl, NUIHttpCacheResponseEntity());
  }

  NUIHttpResponse<D> _processResponse<D>(http.Response response, NUIEntMapper entity){
    _logWeb("Response body : " + response?.body);
    if(response.statusCode == HttpStatus.ok){
      if(entity != null){
        return NUIHttpResponse<D>(
            success: true,
            code: response.statusCode,
            message: _statusMessages[response.statusCode] ?? "Http status OK",
            body: entity.toEntity(jsonDecode(response.body)),
            responseBody: jsonDecode(response.body),
            response: response.body,
            headers: response.headers
        );
      }
      return NUIHttpResponse<D>(
          success: true,
          code: response.statusCode,
          message: _statusMessages[response.statusCode] ?? "Http status OK",
          responseBody: jsonDecode(response.body),
          response: response.body,
          headers: response.headers
      );
    }
    else {
      _processError(response.statusCode?.toString());
      return NUIHttpResponse<D>(
          success: false,
          code: response.statusCode,
          message: _statusMessages[response.statusCode] ?? "No message specified",
          response: response.body,
          responseBody: jsonDecode(response.body),
          headers: response.headers
      );
    }
  }

  void _processError(String code){
    if(_errorHandler != null){
      final context = NUINavigator.getLastContext();
      if(context != null) {
        _errorHandler.processError(context, code, errorType: NUIErrorType.HTTP);
      }
    }
  }

  Future<NUIHttpResponse<T>> _preRequestChecking<T extends NUIEnt>() async{
    final hasConnection = await _checkInternetConnection();
    if(hasConnection) {
      return null;
    }
    else{
      _processError(NUIErrorCode.NO_INTERNET_CONNECTION.value().toString());
      return NUIHttpResponse<T>(success: false, code: NUIErrorCode.NO_INTERNET_CONNECTION.value(), message: "No internet connection found");
    }
  }

  @override
  Future<NUIHttpResponse<T>> doPOST<T extends NUIEnt>({String url, Map<String, String> headers, NUIEnt bodyEnt, Map<String, dynamic> body, Encoding encoding, String pathVariable, NUIEntMapper<T> entity, bool storeTemp}) async{
    final preprocess = await _preRequestChecking<T>();
    if(preprocess != null) return preprocess;
    final targetUrl = _getProcessedUrl(url, pathVariable, null);
    NUIHttpResponse<T> response = await wrapException<T>(() async{
      http.Response response = await http.post(
          targetUrl,
          headers: _getHeaders(headers),
          body: bodyEnt != null ? jsonEncode(bodyEnt) : body,
          encoding: encoding
      );
      if(storeTemp) _storeCacheResponse(url: targetUrl, pathVariable: pathVariable, response: response);
      return _processResponse<T>(response, entity);
    });
    _preHandle<T>(response);
    _postHandle(response);
    return response;
  }

  @override
  Future<NUIHttpResponse<T>> doPUT<T extends NUIEnt>({String url, Map<String, String> headers, NUIEnt bodyEnt, Map<String, dynamic> body, Encoding encoding, String pathVariable, NUIEntMapper<T> entity, bool storeTemp}) async{
    final preprocess = await _preRequestChecking<T>();
    if(preprocess != null) return preprocess;
    final targetUrl = _getProcessedUrl(url, pathVariable, null);
    NUIHttpResponse<T> response = await wrapException<T>(() async{
      http.Response response = await http.put(
          _getProcessedUrl(url, pathVariable, null),
          headers: _getHeaders(headers),
          body: bodyEnt != null ? jsonEncode(bodyEnt) : body,
          encoding: encoding
      );
      if(storeTemp) _storeCacheResponse(url: targetUrl, pathVariable: pathVariable, response: response);
      return _processResponse<T>(response, entity);
    });
    _preHandle<T>(response);
    _postHandle(response);
    return response;
  }

  @override
  Future<NUIHttpResponse<T>> doPATCH<T extends NUIEnt>({String url, Map<String, String> headers, NUIEnt bodyEnt, Map<String, dynamic> body, Encoding encoding, String pathVariable, NUIEntMapper<T> entity, bool storeTemp}) async{
    final preprocess = await _preRequestChecking<T>();
    if(preprocess != null) return preprocess;
    final targetUrl = _getProcessedUrl(url, pathVariable, null);
    NUIHttpResponse<T> response = await wrapException<T>(() async{
      http.Response response = await http.patch(
          _getProcessedUrl(url, pathVariable, null),
          headers: _getHeaders(headers),
          body: bodyEnt != null ? jsonEncode(bodyEnt) : body,
          encoding: encoding
      );
      if(storeTemp) _storeCacheResponse(url: targetUrl, pathVariable: pathVariable, response: response);
      return _processResponse<T>(response, entity);
    });
    _preHandle<T>(response);
    _postHandle(response);
    return response;
  }

  @override
  Future<NUIHttpResponse<T>> doDELETE<T extends NUIEnt>({String url, Map<String, String> headers, Encoding encoding, String pathVariable, NUIEntMapper<T> entity, bool storeTemp}) async{
    final preprocess = await _preRequestChecking<T>();
    if(preprocess != null) return preprocess;
    final targetUrl = _getProcessedUrl(url, pathVariable, null);
    NUIHttpResponse<T> response = await wrapException<T>(() async{
      http.Response response = await http.delete(
          _getProcessedUrl(url, pathVariable, null),
          headers: _getHeaders(headers)
      );
      if(storeTemp) _storeCacheResponse(url: targetUrl, pathVariable: pathVariable, response: response);
      return _processResponse<T>(response, entity);
    });
    _preHandle<T>(response);
    _postHandle(response);
    return response;
  }

  @override
  Future<NUIHttpResponse<T>> doGET<T extends NUIEnt>({String url, Map<String, String> headers, String pathVariable, Map<String, String> params, NUIEntMapper<T> entity, bool storeTemp}) async{
    final preprocess = await _preRequestChecking<T>();
    if(preprocess != null) return preprocess;
    final targetUrl = _getProcessedUrl(url, pathVariable, null);
    NUIHttpResponse<T> response = await wrapException<T>(() async{
      http.Response response = await http.get(
          _getProcessedUrl(url, pathVariable, params),
          headers: _getHeaders(headers),
      );
      if(storeTemp == true) _storeCacheResponse(url: targetUrl, pathVariable: pathVariable, response: response);
      return _processResponse<T>(response, entity);
    });
    _preHandle<T>(response);
    _postHandle(response);
    return response;
  }
}

class NUIWebBuilder{
  Map<int, String> _statusMessages;
  String _baseUrl;
  HttpHeaderSetter _headers;
  String _module;
  HttpAdditionalHandler _preHandler;
  HttpAdditionalHandler _postHandler;
  Duration _postHandlerDelay;
  ConnectivityChangeListener _connectivityListener;
  ErrorHandlerGetter _errorHandlerBuilder;
  NUIErrorHandler _errorHandler;

  NUIWebBuilder module(String module){
    _module = module;
    return this;
  }
  NUIWebBuilder errorHandler(ErrorHandlerGetter errorHandler){
    _errorHandlerBuilder = errorHandler;
    return this;
  }
  NUIWebBuilder preHandler(HttpAdditionalHandler handler){
    _preHandler = handler;
    return this;
  }
  NUIWebBuilder postHandler(HttpAdditionalHandler handler, {Duration delay}){
    _postHandler = handler;
    _postHandlerDelay = delay ?? Duration(milliseconds: 500);
    return this;
  }
  NUIWebBuilder connectivityListener(ConnectivityChangeListener listener){
    this._connectivityListener = listener;
    return this;
  }
  NUIWebBuilder statusMessages(Map<int, String> messages){
    _statusMessages = messages;
    return this;
  }
  NUIWebBuilder baseURL(String url){
    _baseUrl = url;
    return this;
  }
  NUIWebBuilder headers(HttpHeaderSetter headers){
    _headers = headers;
    return this;
  }
  Future<NUIWeb> build() async{
    if(_errorHandlerBuilder != null){
      _errorHandler = await _errorHandlerBuilder().build();
    }
    final web = NUIWeb._builder(this);
    logNUI("NUIWeb", "Initialization successful");
    return web;
  }
}

==> nui_web-0.0.4.tar.gz: lib/nui_web.dart <==
library nui_web;
import 'dart:convert';
import 'dart:developer';
import 'dart:io';
import 'dart:typed_data';
import 'package:flutter/cupertino.dart';
import 'package:http/http.dart' as http;
import 'package:nui_core/busevent/nui_busevent.dart';
import 'package:nui_core/code/nui_code_util.dart';
import 'package:nui_core/data/nui_data_util.dart';
import 'package:nui_core/data/nui_ent.dart';
import 'package:nui_core/nui_core.dart';
import 'package:nui_error_handler/nui_error_handler.dart';
import 'package:nui_web/nui_connection.dart';
import 'nui_web_models.dart';

export 'package:nui_web/nui_web.dart';
export 'package:nui_web/nui_web_models.dart';
export 'package:nui_web/nui_connection.dart';

typedef HttpHeaderSetter = Map<String, String> Function();
typedef HttpExceptionWrapper = Future<NUIHttpResponse> Function();
typedef HttpAdditionalHandler = void Function(NUIHttpResponse);

abstract class NUIWebInterface{
  Future<NUIHttpResponse<T>> doPOST<T extends NUIEnt>({@required String url, Map<String, String> headers, NUIEnt bodyEnt, Map<String, dynamic> body, Encoding encoding, String pathVariable, NUIEntMapper<T> entity, bool storeTemp});
  Future<NUIHttpResponse<T>> doPUT<T extends NUIEnt>({@required String url, Map<String, String> headers, NUIEnt bodyEnt, Map<String, dynamic> body, Encoding encoding, String pathVariable, NUIEntMapper<T> entity, bool storeTemp});
  Future<NUIHttpResponse<T>> doPATCH<T extends NUIEnt>({@required String url, Map<String, String> headers, NUIEnt bodyEnt, Map<String, dynamic> body, Encoding encoding, String pathVariable, NUIEntMapper<T> entity, bool storeTemp});
  Future<NUIHttpResponse<T>> doDELETE<T extends NUIEnt>({@required String url, Map<String, String> headers, Encoding encoding, String pathVariable, NUIEntMapper<T> entity, bool storeTemp});
  Future<NUIHttpResponse<T>> doGET<T extends NUIEnt>({@required String url, Map<String, String> headers, String pathVariable, Map<String, String> params, NUIEntMapper<T> entity, bool storeTemp});
}

class NUIWeb implements NUIWebInterface{
  static const _NUIWEB_MODULE = "NUIWeb";

  NUIWeb._builder(NUIWebBuilder builder) : _baseUrl = builder._baseUrl, _preHandler = builder._preHandler, _postHandlerDelay = builder._postHandlerDelay,  _postHandler = builder._postHandler, _headers = builder._headers, _statusMessages = builder._statusMessages ?? Map(), _errorHandler = builder._errorHandler, _connectivityListener = builder._connectivityListener, _module = isNullOrEmpty(builder._module) ? _defaultModule : builder._module{
    _webMap[_module] = this;
    NUISharedPrefBuilder().build();
    NUIBusEventBuilder().module(_getBusEventModule()).build();
    NUIConnectionUtil.subscribeConnectivityChange((result, hasInternet) {
      _getBusEvent().send<bool>(NUIBusEventData<bool>(data: hasInternet, sender: _getBusEventModule(), code: NUIConnectionUtil.CONNECTIVITY_CHANNEL), channel: NUIConnectionUtil.CONNECTIVITY_CHANNEL);
      if(_connectivityListener != null) {
        _connectivityListener(result, hasInternet);
      }
    });
  }

  final Map<int, String> _statusMessages;
  final String _module;
  final String _baseUrl;
  final HttpAdditionalHandler _preHandler;
  final HttpAdditionalHandler _postHandler;
  final Duration _postHandlerDelay;
  final ConnectivityChangeListener _connectivityListener;
  final HttpHeaderSetter _headers;
  final NUIErrorHandler _errorHandler;
  static Map<String, NUIWeb> _webMap = Map();
  static const String _defaultModule = "NUIWebDefaultModule";

  static Future<Uint8List> downloadFromURL(String url) async{
    try {
      http.Client client = new http.Client();
      var req = await client.get(Uri.parse(url));
      var bytes = req.bodyBytes;
      return bytes;
    }catch(e){
      return null;
    }
  }

  static NUIWeb get({String module}){
    NUIWeb instance = _webMap[isNullOrEmpty(module) ? _defaultModule : module];
    if(instance == null) throw Exception("NUIWeb -> NUIWeb instance has not been initialized with builder");
    return instance;
  }

  NUIBusEvent _getBusEvent(){
    return NUIBusEvent.get(module: _getBusEventModule());
  }

  void sendBusEvent<T>(NUIBusEventData<T> data, {String channel: NUIConnectionUtil.CONNECTIVITY_CHANNEL}){
    _getBusEvent().send<T>(data, channel: channel);
  }

  void subscribeBusEvent<T>(NUIBusEventListener listener, {String channel: NUIConnectionUtil.CONNECTIVITY_CHANNEL}){
    _getBusEvent().listen(listener, channel: channel);
  }

  void disposeBusEvent(NUIBusEventListener listener, {String channel: NUIConnectionUtil.CONNECTIVITY_CHANNEL}){
    _getBusEvent().dispose(listener, channel: channel);
  }

  String _getBusEventModule(){
    return "NUIWebBusEvent_" + (_module ?? _defaultModule);
  }

  String _getProcessedUrl(String url, String pathVariable, Map<String, String> params){
    String webUrl = isNullOrEmpty(_baseUrl) ? url : (_baseUrl[_baseUrl.length - 1] != '/' ? (_baseUrl + "/" + url) : (_baseUrl + url));
    if(!isNullOrEmpty(pathVariable)){
      if(webUrl[webUrl.length - 1] != '/') {
        webUrl = webUrl + "/" + pathVariable;
      }
      else{
        webUrl = webUrl + pathVariable;
      }
    }
    webUrl =  webUrl + _getPathParams(params);
    logNUI(NUIWeb._NUIWEB_MODULE, "URL : $webUrl");
    return webUrl;
  }

  void _logWeb(String message){
    logNUI(NUIWeb._NUIWEB_MODULE, "$message");
  }

  void _preHandle<T>(NUIHttpResponse<T> response){
    if(_preHandler != null) _preHandler(response);
  }

  void _postHandle<T>(NUIHttpResponse<T> response){
    if(_postHandler != null){
      NUIAsync.delay(_postHandlerDelay, () {
        _postHandler(response);
      });
    }
  }

  String _getPathParams(Map<String, String> params){
    if(params == null || params.isEmpty){
      return "";
    }
    String pathParam = "";
    for(String key in params.keys){
      pathParam = pathParam + (pathParam.isEmpty ? "?$key=${params[key]}" : "&$key=${params[key]}");
    }
    return pathParam;
  }

  Map<String, String> _getHeaders(Map<String, String> additionalHeaders){
    if(_headers == null) return additionalHeaders;
    if(additionalHeaders == null) return _headers();
    Map<String, String> headers = _headers();
    for(String key in additionalHeaders.keys){
      headers[key] = additionalHeaders[key];
    }
    if(headers.isNotEmpty){
      _logWeb("Headers : " + headers.toString());
    }
    return headers;
  }

  Future<NUIHttpResponse<T>> wrapException<T extends NUIEnt>(HttpExceptionWrapper task) async{
    try {
      return task();
    }catch(e){
      _logWeb("Encountered an exception : ${e.toString()}");
      return NUIHttpResponse(success: false, code: HttpStatus.internalServerError, message: e.toString());
    }
  }

  Future<bool> _checkInternetConnection() async{
    return NUIConnectionUtil.hasConnection();
  }

  void _storeCacheResponse({String url, String pathVariable, http.Response response}){
    if(response.statusCode != HttpStatus.ok) return;
    final fullUrl = _getProcessedUrl(url, pathVariable, null);
    final cacheResponse = NUIHttpCacheResponse(response: response.body, headers: response.headers);
    NUISharedPref.get().save(fullUrl, cacheResponse);
  }

  NUIHttpCacheResponse getCacheResponse({String url, String pathVariable}){
    final fullUrl = _getProcessedUrl(url, pathVariable, null);
    return NUISharedPref.get().quickGetObject(fullUrl, NUIHttpCacheResponseEntity());
  }

  NUIHttpResponse<D> _processResponse<D>(http.Response response, NUIEntMapper entity){
    _logWeb("Response body : " + response?.body);
    if(response.statusCode == HttpStatus.ok){
      if(entity != null){
        return NUIHttpResponse<D>(
            success: true,
            code: response.statusCode,
            message: _statusMessages[response.statusCode] ?? "Http status OK",
            body: entity.toEntity(jsonDecode(response.body)),
            responseBody: jsonDecode(response.body),
            response: response.body,
            headers: response.headers
        );
      }
      return NUIHttpResponse<D>(
          success: true,
          code: response.statusCode,
          message: _statusMessages[response.statusCode] ?? "Http status OK",
          responseBody: jsonDecode(response.body),
          response: response.body,
          headers: response.headers
      );
    }
    else {
      _processError(response.statusCode?.toString());
      return NUIHttpResponse<D>(
          success: false,
          code: response.statusCode,
          message: _statusMessages[response.statusCode] ?? "No message specified",
          response: response.body,
          responseBody: jsonDecode(response.body),
          headers: response.headers
      );
    }
  }

  void _processError(String code){
    if(_errorHandler != null){
      final context = NUINavigator.getLastContext();
      if(context != null) {
        _errorHandler.processError(context, code, errorType: NUIErrorType.HTTP);
      }
    }
  }

  Future<NUIHttpResponse<T>> _preRequestChecking<T extends NUIEnt>() async{
    final hasConnection = await _checkInternetConnection();
    if(hasConnection) {
      return null;
    }
    else{
      _processError(NUIErrorCode.NO_INTERNET_CONNECTION.value().toString());
      return NUIHttpResponse<T>(success: false, code: NUIErrorCode.NO_INTERNET_CONNECTION.value(), message: "No internet connection found");
    }
  }

  @override
  Future<NUIHttpResponse<T>> doPOST<T extends NUIEnt>({String url, Map<String, String> headers, NUIEnt bodyEnt, Map<String, dynamic> body, Encoding encoding, String pathVariable, NUIEntMapper<T> entity, bool storeTemp: false}) async{
    final preprocess = await _preRequestChecking<T>();
    if(preprocess != null) return preprocess;
    final targetUrl = _getProcessedUrl(url, pathVariable, null);
    NUIHttpResponse<T> response = await wrapException<T>(() async{
      http.Response response = await http.post(
          targetUrl,
          headers: _getHeaders(headers),
          body: bodyEnt != null ? jsonEncode(bodyEnt) : body,
          encoding: encoding
      );
      if(storeTemp == true) _storeCacheResponse(url: targetUrl, pathVariable: pathVariable, response: response);
      return _processResponse<T>(response, entity);
    });
    _preHandle<T>(response);
    _postHandle(response);
    return response;
  }

  @override
  Future<NUIHttpResponse<T>> doPUT<T extends NUIEnt>({String url, Map<String, String> headers, NUIEnt bodyEnt, Map<String, dynamic> body, Encoding encoding, String pathVariable, NUIEntMapper<T> entity, bool storeTemp: false}) async{
    final preprocess = await _preRequestChecking<T>();
    if(preprocess != null) return preprocess;
    final targetUrl = _getProcessedUrl(url, pathVariable, null);
    NUIHttpResponse<T> response = await wrapException<T>(() async{
      http.Response response = await http.put(
          _getProcessedUrl(url, pathVariable, null),
          headers: _getHeaders(headers),
          body: bodyEnt != null ? jsonEncode(bodyEnt) : body,
          encoding: encoding
      );
      if(storeTemp == true) _storeCacheResponse(url: targetUrl, pathVariable: pathVariable, response: response);
      return _processResponse<T>(response, entity);
    });
    _preHandle<T>(response);
    _postHandle(response);
    return response;
  }

  @override
  Future<NUIHttpResponse<T>> doPATCH<T extends NUIEnt>({String url, Map<String, String> headers, NUIEnt bodyEnt, Map<String, dynamic> body, Encoding encoding, String pathVariable, NUIEntMapper<T> entity, bool storeTemp: false}) async{
    final preprocess = await _preRequestChecking<T>();
    if(preprocess != null) return preprocess;
    final targetUrl = _getProcessedUrl(url, pathVariable, null);
    NUIHttpResponse<T> response = await wrapException<T>(() async{
      http.Response response = await http.patch(
          _getProcessedUrl(url, pathVariable, null),
          headers: _getHeaders(headers),
          body: bodyEnt != null ? jsonEncode(bodyEnt) : body,
          encoding: encoding
      );
      if(storeTemp == true) _storeCacheResponse(url: targetUrl, pathVariable: pathVariable, response: response);
      return _processResponse<T>(response, entity);
    });
    _preHandle<T>(response);
    _postHandle(response);
    return response;
  }

  @override
  Future<NUIHttpResponse<T>> doDELETE<T extends NUIEnt>({String url, Map<String, String> headers, Encoding encoding, String pathVariable, NUIEntMapper<T> entity, bool storeTemp: false}) async{
    final preprocess = await _preRequestChecking<T>();
    if(preprocess != null) return preprocess;
    final targetUrl = _getProcessedUrl(url, pathVariable, null);
    NUIHttpResponse<T> response = await wrapException<T>(() async{
      http.Response response = await http.delete(
          _getProcessedUrl(url, pathVariable, null),
          headers: _getHeaders(headers)
      );
      if(storeTemp == true) _storeCacheResponse(url: targetUrl, pathVariable: pathVariable, response: response);
      return _processResponse<T>(response, entity);
    });
    _preHandle<T>(response);
    _postHandle(response);
    return response;
  }

  @override
  Future<NUIHttpResponse<T>> doGET<T extends NUIEnt>({String url, Map<String, String> headers, String pathVariable, Map<String, String> params, NUIEntMapper<T> entity, bool storeTemp}) async{
    final preprocess = await _preRequestChecking<T>();
    if(preprocess != null) return preprocess;
    final targetUrl = _getProcessedUrl(url, pathVariable, null);
    NUIHttpResponse<T> response = await wrapException<T>(() async{
      http.Response response = await http.get(
          _getProcessedUrl(url, pathVariable, params),
          headers: _getHeaders(headers),
      );
      if(storeTemp == true) _storeCacheResponse(url: targetUrl, pathVariable: pathVariable, response: response);
      return _processResponse<T>(response, entity);
    });
    _preHandle<T>(response);
    _postHandle(response);
    return response;
  }
}

class NUIWebBuilder{
  Map<int, String> _statusMessages;
  String _baseUrl;
  HttpHeaderSetter _headers;
  String _module;
  HttpAdditionalHandler _preHandler;
  HttpAdditionalHandler _postHandler;
  Duration _postHandlerDelay;
  ConnectivityChangeListener _connectivityListener;
  ErrorHandlerGetter _errorHandlerBuilder;
  NUIErrorHandler _errorHandler;

  NUIWebBuilder module(String module){
    _module = module;
    return this;
  }
  NUIWebBuilder errorHandler(ErrorHandlerGetter errorHandler){
    _errorHandlerBuilder = errorHandler;
    return this;
  }
  NUIWebBuilder preHandler(HttpAdditionalHandler handler){
    _preHandler = handler;
    return this;
  }
  NUIWebBuilder postHandler(HttpAdditionalHandler handler, {Duration delay}){
    _postHandler = handler;
    _postHandlerDelay = delay ?? Duration(milliseconds: 500);
    return this;
  }
  NUIWebBuilder connectivityListener(ConnectivityChangeListener listener){
    this._connectivityListener = listener;
    return this;
  }
  NUIWebBuilder statusMessages(Map<int, String> messages){
    _statusMessages = messages;
    return this;
  }
  NUIWebBuilder baseURL(String url){
    _baseUrl = url;
    return this;
  }
  NUIWebBuilder headers(HttpHeaderSetter headers){
    _headers = headers;
    return this;
  }
  Future<NUIWeb> build() async{
    if(_errorHandlerBuilder != null){
      _errorHandler = await _errorHandlerBuilder().build();
    }
    final web = NUIWeb._builder(this);
    logNUI(NUIWeb._NUIWEB_MODULE, "Initialization successful");
    return web;
  }
}
